testGeneCategories.regression<-function(...){
  testGeneCategories(method="poissonregression", ...)
}

testGeneCategories.dircomp<-function(...){
  testGeneCategories(method="poissongroups", ...)
}

testGeneCategories<-function(data, method="GaussianSum", category="cogpathways", samples=c(1, 2), multtest="BH", ord=TRUE,
                             enrichment.method="Binomial", enrichment.p=0.1,
			     design.matrix=NULL, coef=1, covariates=NULL, log.covariates=FALSE, groups=NULL,
			     overdisp=FALSE){			    

  Categories<-getCategory(category, data$Type)
  
  if(is.null(Categories)){
    stop("No category specified.")
  }

  Statistics<-list()
  Statistics$Type<-data$Type
  Statistics$Category<-category
 
  if (tolower(method)=="gaussiansum"){
  
    res.stat<-testGeneCategories.gaussiansum(data, Categories, samples=samples)
    Statistics$RankingType<-"GaussianSum"
    
  } else if (tolower(method)=="independence"){
  
    res.stat<-testGeneCategories.independence(data, Categories, samples=samples)
    Statistics$RankingType<-"Independence"
    
  } else if (tolower(method)=="enrichment"){
  
    if(missing(enrichment.method) | missing(enrichment.p)){
      stop("Both enrichment.method and enrichment.p needs to be specified when using the enrichment method.")
    }
  
    res.stat<-testGeneCategories.enrichment(data, Categories, samples=samples, alpha=enrichment.p, method=enrichment.method)
    Statistics$RankingType<-sprintf("Enrichment, p-value cutoff=%s", enrichment.p)
    
  } else if (tolower(method)=="poisson"){
  
    if(is.null(design.matrix)){
      stop(sprintf("Error! Design matrix necessary for the general Poisson model.\n"))
    }  
    
    res.stat<-testGeneCategories.poisson(data, Categories, design.matrix, coef, overdisp=overdisp)
    Statistics$RankingType<-"Poisson"
    Statistics$design.matrix<-design.matrix
    Statisitcs$coef<-coef
    
  } else if (tolower(method)=="poissonregression"){
  
    if(is.null(covariates)){    
      stop(sprintf("Error! Covariates necessary for Poisson regression model.\n"))
    }
        
    Statistics$RankingType<-"PoissonRegression"
    Statistics$coef<-coef
  
    if(log.covariates==TRUE){
      Statistics$covariates<-log10(covariates)
      Statistics$covariates.antilog<-covariates
    } else { 
      Statistics$covariates<-covariates
    }
  
    design.matrix<-matrix(Statistics$covariates, nc=1)
    Statistics$design.matrix<-design.matrix
 
    res.stat<-testGeneCategories.poisson(data, Categories, design.matrix, coef=1, overdisp=overdisp)
    
  } else if (tolower(method)=="poissongroups"){
  
    if(is.null(groups)){
      stop(sprintf("Error! Groups necessary for Poisson direction comparison.\n"))
    }
    
    if(length(unique(groups))!=2){
      stop(sprintf("Error! Only two groups allowed in Poisson direct comparison.\n"))
    }
  
    Statistics$RankingType<-"PoissonGroups"
    Statistics$coef<-1
    Statistics$groups<-groups
    design.matrix<-matrix(as.integer(Statistics$groups==(unique(Statistics$groups)[1])), nc=1)
    Statistics$design.matrix<-design.matrix

    res.stat<-testGeneCategories.poisson(data, Categories, Statistics$design.matrix, coef=Statistics$coef, overdisp=overdisp)
    
  } else { 
    stop(sprintf("Unknown method %s.\n", method))
  }     

  Statistics$Ranking.details<-res.stat$Ranking.details

  CatStat<-generalCatStatistics(data, Categories)

  Statistics$Ranking<-data.frame(CatStat, res.stat$PValues, names(Categories))
  colnames(Statistics$Ranking)<-c(colnames(CatStat), colnames(res.stat$PValues), "Category")
  
  
  if(!is.null(multtest)){
  
    # Call multtest correction procedure
    multtest.res<-mt.rawp2adjp(Statistics$Ranking[,"P-value"], proc=multtest)
 
    # Return original order
    multtest.p<-multtest.res$adjp[order(multtest.res$index),2]
         
    # Save information regarding multiple testing procedue
    Statistics$Multtest<-multtest

   # Create a new data.frame
    Statistics$Ranking<-data.frame(CatStat, res.stat$PValues, multtest.p, names(Categories))
    colnames(Statistics$Ranking)<-c(colnames(CatStat), colnames(res.stat$PValues), sprintf("Adjusted P-value (%s)", multtest), "Category")
   
  }
  
  # Add data and annotation
  Statistics$Data<-data$Data
  Statistics$Annotation<-data$Annotation
  
  
  # Sort list
  if(ord==TRUE){
    ord.ranking<-order(Statistics$Ranking[,"P-value"])
    Statistics$Ranking<-Statistics$Ranking[ord.ranking,]  
  }
  
  return(Statistics)
}


testGeneCategories.gaussiansum<-function(data, categories, samples){

  nCat<-length(categories)

  stat<-testGeneFamilies(data, method="gaussian", samples=samples)
  
  PValues<-rep(NA, nCat)
  CatStat<-rep(NA, nCat)
  Zscore<-rep(NA, nCat)
  CatSize<-numeric(length=nCat)
  CatSizeTotal<-numeric(length=nCat)
  
  for(i in 1:nCat){

    CatSizeTotal[i]<-length(categories[[i]])
    genesincat<-stat$Ranking[,1]%in%categories[[i]]
    CatSize[i]<-sum(genesincat)
    
    if(CatSize[i]>0){
    
      CatStat[i]<-mean(stat$Ranking[genesincat,2])
      stderr<-sqrt(1/CatSize[i])
      Zscore[i]<-CatStat[i]/stderr
      PValues[i]<-(1-pnorm(abs(CatStat[i]), sd=stderr))*2
    }
  
  }
  
  Result<-data.frame(Zscore, PValues)
  colnames(Result)<-c("Gaussian sum (Z-score)", "P-value")
  
  Details<-list()
  Details$data<-data
  Details$stdev<-stderr
  Details$samples<-samples
  
  return(list(PValues=Result, Ranking.details=Details))
}

testGeneCategories.independence<-function(data, categories, samples, simulate.p.value=TRUE, B=2000){

  nCat<-length(categories)
    
  PValues<-rep(NA, nCat)
  CatSize<-numeric(length=nCat)
  CatSizeTotal<-numeric(length=nCat)
  
  for(i in 1:nCat){

    CatSizeTotal[i]<-length(categories[[i]])
    genesincat<-data$Data[,1]%in%categories[[i]]
    CatSize[i]<-sum(genesincat)
    if(CatSize[i]>0){    
      counts.table<-rbind(data$Data[genesincat,samples], apply(data$Data[!genesincat, samples], 2, sum))
      PValues[i]<-fisher.test(counts.table, simulate.p.value=simulate.p.value, B=B)$p.value
    }
  }
  
  Result<-data.frame(PValues)
  colnames(Result)<-c("P-value")
  
  Details<-list()
  Details$data<-data
  Details$B<-B
  Details$simulate.p.value<-simulate.p.value
  Details$samples<-samples
  
  return(list(PValues=Result, Ranking.details=Details))
}

testGeneCategories.enrichment<-function(data, categories, samples, method, alpha){

  nCat<-length(categories)
  stat<-testGeneFamilies(data, method=method, samples=samples, ord=TRUE, multtest="BH")  
  
  top.genes<-stat$Ranking[,"P-value"]<=alpha
  ntop.genes<-sum(top.genes)
    
  PValues<-rep(NA, nCat)
  CatStat<-rep(NA, nCat)
  CatSize<-numeric(length=nCat)
  CatSizeTotal<-numeric(length=nCat)
  n<-matrix(NA, nc=4, nr=nCat)
  
  for(i in 1:nCat){

    CatSizeTotal[i]<-length(categories[[i]])
    genesincat<-data$Annotation[,1]%in%categories[[i]]
    CatSize[i]<-sum(genesincat)

    if(CatSize[i]>0){

      n1<-sum(top.genes&genesincat)
      n2<-sum(top.genes&!genesincat)
      n3<-sum(!top.genes&genesincat)
      n4<-sum(!top.genes&!genesincat)
        
      mat<-matrix(c(n1, n2, n3, n4), byrow=TRUE, nc=2)
      PValues[i]<-fisher.test(mat, alternative="two.sided")$p.value
      n[i,]<-c(n1, n2, n3, n4)
    }
    
  }
  
  Result<-data.frame(n, PValues)
  colnames(Result)<-c("Sig. gene families in cat.", "Sig. gene families not in cat.", "Non-sig. gene families in cat", "Non-sig. gene families not in cat", "P-value")
  
  Details<-list()
  Details$data<-data
  Details$samples<-samples
  Details$method<-method
  Details$alpha<-alpha
  
  return(list(PValues=Result, Ranking.details=Details))
}

testGeneCategories.poisson<-function(data, categories, design.matrix, coef, overdisp=FALSE){
  
  nCat<-length(categories) 
  n.coef<-dim(design.matrix)[2]
  nrows<-dim(data$Data)[2]

  CatSize<-numeric(length=nCat)
  CatSizeTotal<-numeric(length=nCat) 
  PValues<-rep(NA, nCat)
  Coef<-rep(NA, nCat)
  R2<-rep(NA, nCat)
  AIC<-rep(NA, nCat)
  
  TotalCounts<-apply(data$Data,2,sum)

  for(i in 1:nCat){

    CatSizeTotal[i]<-length(categories[[i]])    
    genesincat<-data$Annotation[,1]%in%categories[[i]]
    CatSize[i]<-sum(genesincat)

    if(CatSize[i]>1){
      CountsCat<-data$Data[genesincat,]
      GeneFamilies<-as.character(data$Annotation[genesincat,1])
        
      res.po<-testGeneCategories.poisson.fit(CountsCat, TotalCounts, GeneFamilies, design.matrix, overdisp=overdisp)
	
      # Get p-value and value of the coefficient
      PValues[i]<-summary(res.po$fit)$coef[paste("Coef", coef, sep=""),4]
      Coef[i]<-summary(res.po$fit)$coef[paste("Coef", coef, sep=""),1]
      R2[i]<-res.po$R2
      AIC[i]<-res.po$AIC
    }
    
  }
  
  Result<-data.frame(Coef, AIC, PValues)
  colnames(Result)<-c("Coefficient", "AIC", "P-value")
  
  Details<-list()
  Details$data<-data
  Details$design.matrix<-design.matrix  
  Details$R2<-R2
  Details$TotalCounts<-TotalCounts
  Details$AIC<-AIC

  return(list(PValues=Result, Ranking.details=Details))
   
}

testGeneCategories.poisson.fit<-function(counts, counts.total, gene.families, design.matrix, overdisp=FALSE){
    
  n<-length(gene.families)
  m<-dim(counts)[2]
  n.coef<-dim(design.matrix)[2]
  
  if(overdisp==FALSE){
    family.glm<-poisson
  } else { 
    family.glm<-quasipoisson
  }
  
  data.mat<-data.frame(as.numeric(as.matrix(counts)), rep(counts.total, each=n), 
                       rep(as.character(gene.families), m), matrix(rep(design.matrix, each=n), nc=n.coef))
  colnames(data.mat)<-c("Counts", "Total", "COG", paste("Coef", 1:n.coef, sep=""))

  form<-paste("Counts ~ 0 + offset(log(Total))+ COG + ", paste(paste("Coef", 1:n.coef, sep=""), collapse=" + "), sep="")
  res.glm<-glm(formula(form), data=data.mat, family=family.glm)
  
  
  Result<-list()  
  Result$fit<-res.glm
  Result$n.coef<-n.coef
  Result$data.mat<-data.mat
  Result$form<-form
  Result$R2<-calculatePoissonR2(res.glm, k=n.coef-1)
  Result$AIC<-res.glm$aic
  
  return(Result)

}

generalCatStatistics<-function(Data, categories){

  nCat<-length(categories)
  
  Found<-numeric(length=nCat)
  Size<-numeric(length=nCat)
  
  
  for(i in 1:nCat){
  
    Size[i]<-length(categories[[i]])
    Found[i]<-sum(Data$Annotation[,1]%in%categories[[i]])
  }
  
  Result<-data.frame(Size, Found)
  colnames(Result)<-c("Category size", "Gene families found")

  return(Result)
 
}


getCategory<-function(category, type){

  if(tolower(category)=="cogpathways"){
    if(type!="COG"){
      stop(sprintf("Wrong datatype! Expected COG, found %s.\n", type))
    }    
    categories<-ShotgunFunctionalizeRAnnotationData$PATHWAYS$COG$PATHWAY2COG
  } else if (tolower(category)=="cogcategories"){
    if(type!="COG"){
      stop(sprintf("Wrong datatype! Expected COG, found %s.\n", type))
    }    
    categories<-ShotgunFunctionalizeRAnnotationData$PATHWAYS$COG$CATEGORY2COG
  } else if (tolower(category)=="cogphyla"){
    if(type!="COG"){
      stop(sprintf("Wrong datatype! Expected COG, found %s.\n", type))
    }    
    categories<-ShotgunFunctionalizeRAnnotationData$PATHWAYS$COG$PHYLA2COG 
  } else if(tolower(category)=="kegg"){
    if(type!="EC"){
      stop(sprintf("Wrong datatype! Expected EC, found %s.\n", type))
    }    
    categories<-ShotgunFunctionalizeRAnnotationData$PATHWAYS$KEGG$PATHWAY2EC
  } else {
    stop(sprintf("Unknown category %s.\n", category))
  }

  return(categories)

}
